module OMD
  VERSION="0.1"
  
  # All loaded Packages, hashed by Package id
  PACKAGES = { }
  
  # All loaded Classes, hashed by Class name
  CLASSES  = { }
  
  # Takes an XML string using the ObjectModelDescription schema.
  # Returns an array of OMD::Packages from the file.
  def self.load( in_xml_string )
    require 'rexml/document'
    doc = REXML::Document.new( in_xml_string, :ignore_whitespace_nodes => :all )
    
    doc.each_element('/ObjectModelDescription/Package'){ |package_element|
      package = Package.from_omd_xml( package_element )
      if PACKAGES[ package.id ]
        warn "Warning: Package '#{package.id}' already loaded" 
      end
      PACKAGES[ package.id ] = package
    }
    
    resolve_references
    
    PACKAGES.values
  end
  
  def self.resolve_references
    CLASSES.each{ |_, omd_class|
      omd_class.parent_classes.map!{ |pclass|
        case pclass
          when String
            CLASSES[ pclass ]
          else
            pclass
        end
      }
      omd_class.parent_classes.compact!
      update_type = lambda{ |omd_item|
        if type = CLASSES[ omd_item.type ]
          omd_item.type = type
        else
          warn "Can't resolve type #{omd_item.type.inspect}"
        end
      }
      omd_class.own_constants.each  &update_type
      omd_class.own_attributes.each &update_type
      omd_class.own_methods.each{ |omd_method|
        omd_method.parameters.each    &update_type
        omd_method.return_values.each &update_type
      }
    }
  end
  
  def self.version_string_value( version_string )
    value = 0
    version_string.split('.').each_with_index{ |s, i|
      value += 100000**(4 - i) * s.to_i
    }
    value
  end
  
  private
    def resolve_type_for( omd_item )
      
    end
end

module OMD::Documentable
  attr_accessor :overview
  attr_accessor :docs
  def add_documentation_from_omd_xml( element )
    if element.attributes[ 'overview' ]
      warn "Warning: Replacing existing overview for #{self.inspect}" if @overview
      @overview = element.attributes[ 'overview' ]
    end
    if docs = element.elements[ 'Documentation' ]
      warn "Warning: Replacing existing documentation for #{self.inspect}" if @docs
      @docs = docs.text
    end
  end
end

module OMD::OwnedByClass
  attr_accessor :owning_class
end

module OMD::OwnedByMethod
  attr_accessor :owning_method
  def owning_class
    @owning_method.owning_class
  end
end

module OMD::DefinedInSingleFile
  attr_reader :source_file
  def find_source_file_from_omd_xml( element )
    walker = element
    walker = walker.parent while walker && !walker.attributes[ 'defined-in' ]
    @source_file = walker && walker.attributes[ 'defined-in' ]
  end
end

class OMD::Package
  include OMD::Documentable
  attr_reader :id, :name, :version
  attr_reader :classes
  def initialize( id, name, version )
    @id = id
    @name = name
    @classes = []
    @version = version
  end
  
  # An array of every file used in this package
  def source_files
    @classes.map{ |_,c| c.source_files }.flatten.compact.uniq
  end
  
  # package_element is an REXML::Element
  def self.from_omd_xml( package_element )
    omd_package = new(
      package_element.attributes[ 'id' ],
      package_element.attributes[ 'name' ],
      package_element.attributes[ 'version' ]
    )
    omd_package.add_documentation_from_omd_xml( package_element )

    #FIXME: What a gross hack this is;
    #       used to simply assign packages to constants/attributes/methods during creation.
    #       should instead walk up the XML tree to find the appropriate package, or something
    $current_package = omd_package    
    
    class_handler = proc{ |class_element|
      omd_class = OMD::Class.from_omd_xml( class_element )
      omd_class.packages << omd_package
      omd_package.classes << omd_class
    }
    package_element.each_element( 'Class', &class_handler )
    package_element.each_element( 'Module', &class_handler )
    package_element.each_element( 'Interface', &class_handler )
    
    omd_package
  end
  
  def constants
    @classes.map{ |c| c.constants  }.flatten.uniq.select{ |k| k.package == self }.sort_by{ |k| k.name }
  end
  
  def attributes
    @classes.map{ |c| c.attributes }.flatten.uniq.select{ |a| a.package == self }.sort_by{ |a| a.name }
  end
  
  def methods
    @classes.map{ |c| c.methods    }.flatten.uniq.select{ |m| m.package == self }.sort_by{ |m| m.name }
  end
  
  def inspect
    "<#{self.class} '#{@name} (#{@id})' :: #{@classes.length} classes>"
  end
end

# Handles <Class> elements as well as <Module> and <Interface> elements.
class OMD::Class
  include OMD::Documentable
  attr_reader :name
  
  # One of :class, :module, or :interface
  attr_reader :type
  
  # An array of OMD::Class instances; may also use a name (for later resolution via OMD.resolve_references).
  # This includes modules and interfaces; use inherits_from( :class ) to find only class parents.
  attr_accessor :parent_classes
  
  # An array of OMD::Packages that own this class.
  attr_reader :packages
  
  # An array of every OMD::Constant in this class.
  attr_reader :own_constants

  # An array of every OMD::Attribute in this class.
  attr_reader :own_attributes

  # An array of every OMD::Method in this class.
  attr_reader :own_methods
  
  # An array of every OMD::Constant that this class inherits.
  attr_reader :inherited_constants

  # An array of every OMD::Attribute that this class inherits.
  attr_reader :inherited_attributes

  # An array of every OMD::Method that this class inherits.
  attr_reader :inherited_methods
  
  # An array of every OMD::Constant available in this class, including inherited constants.
  attr_reader :constants

  # An array of every OMD::Attribute available in this class, including inherited attributes.
  attr_reader :attributes

  # An array of every OMD::Method available in this class, including inherited methods.
  attr_reader :methods

  # An array of every file that defines something in this class
  def source_files
    # FIXME: Doesn't handle the case where the class itself is defined in one or more files separately from any contents
    files1 = @own_constants.map{ |e| e.source_file }
    files2 = @own_attributes.map{ |e| e.source_file }
    files3 = @own_methods.map{ |e| e.source_file }
    files1.concat( files2 ).concat( files3 ).flatten.compact.uniq
  end
  
  def initialize( name, type )
    @name = name
    @type = type
    @packages = []
    @own_constants = []
    @own_attributes = []
    @own_methods = []
    @parent_classes = []
  end
  
  # May re-use another OMD::Class of the same name, if present.
  def self.from_omd_xml( element )
    name = element.attributes[ 'name' ]
    type = element.name.downcase.to_sym
    omd_class = OMD::CLASSES[ name ] ||= new( name, type )
    omd_class.add_documentation_from_omd_xml( element )
    
    # Mixins and interfaces come later (higher precedence) than parent class(es)
    # (Ruby can't have more than one parent class, but other languages can)
    %w| inherits-from mixes-in implements |.each{ |attribute|
      omd_class.parent_classes.concat( ( element.attributes[ attribute ] || "" ).split( /\s+/ ) )
    }
    omd_class.parent_classes.delete( "" )
    omd_class.parent_classes.compact!
    
    element.each_element( 'Constant' ){ |constant_element|
      omd_constant = OMD::Constant.from_omd_xml( constant_element )
      omd_class.own_constants << omd_constant
      omd_constant.owning_class = omd_class
    }
    
    element.each_element( 'Attribute' ){ |attribute_element|
      omd_attribute = OMD::Attribute.from_omd_xml( attribute_element )
      omd_class.own_attributes << omd_attribute
      omd_attribute.owning_class = omd_class
    }
    
    element.each_element( 'Method' ){ |method_element|
      omd_method = OMD::Method.from_omd_xml( method_element )
      omd_class.own_methods << omd_method
      omd_method.owning_class = omd_class
    }
    
    omd_class
  end

  #TODO: DRY these three methods
  def inherited_constants #:nodoc:
    parent_classes.map{ |omd_class| omd_class.constants }.flatten.compact
  end

  def inherited_attributes #:nodoc:
    parent_classes.map{ |omd_class| omd_class.attributes }.flatten.compact
  end

  def inherited_methods #:nodoc:
    parent_classes.map{ |omd_class| omd_class.methods }.flatten.compact
  end

  #TODO: DRY these three methods
  def constants #:nodoc:
    inherited_constants + own_constants
  end

  def attributes #:nodoc:
    inherited_attributes + own_attributes
  end

  def methods #:nodoc:
    inherited_methods + own_methods
  end

  def inherits_from( type=:class )
    @parent_classes.select{ |k| k.type == type }
  end

  def inspect
    "<#{self.class} '#{@name}' :: #{@own_constants.length} constants, #{@own_attributes.length} attributes, #{@own_methods.length} methods>"
  end
end

class OMD::Constant
  include OMD::Documentable
  include OMD::OwnedByClass
  include OMD::DefinedInSingleFile
  attr_reader :name
  attr_accessor :value, :type, :package
  def initialize( name, type=nil, value=nil )
    @name = name
    @type = type
    @value = value
  end
  def self.from_omd_xml( element )
    omd_constant = new(
      element.attributes[ 'name' ],
      element.attributes[ 'type' ],
      element.attributes[ 'value' ]
    )
    omd_constant.add_documentation_from_omd_xml( element )
    omd_constant.find_source_file_from_omd_xml( element )

    omd_constant.package = $current_package
    
    omd_constant
  end

  def inspect
    "<#{self.class} '#{@name}'>"
  end
end

class OMD::Attribute
  include OMD::Documentable
  include OMD::OwnedByClass
  include OMD::DefinedInSingleFile
  attr_reader :name
  attr_accessor :type, :scope, :readwrite, :default, :package
  def initialize( name )
    @name = name
    @readwrite = :all       # schema default
    @scope     = :instance  # schema default
    @access    = :public    # schema default
  end
  def self.from_omd_xml( element )
    omd_attribute = new( element.attributes[ 'name' ] )
    omd_attribute.add_documentation_from_omd_xml( element )
    omd_attribute.find_source_file_from_omd_xml( element )

    omd_attribute.type = element.attributes[ 'type' ]
    omd_attribute.default = element.attributes[ 'default' ]
    if element.attributes[ 'read-write' ]
      omd_attribute.readwrite = element.attributes[ 'read-write' ].to_sym
    end
    if element.attributes[ 'scope' ]
      omd_attribute.scope = element.attributes[ 'scope' ].to_sym
    end
    if element.attributes[ 'access' ]
      omd_attribute.scope = element.attributes[ 'access' ].to_sym
    end

    omd_attribute.package = $current_package
    
    omd_attribute
  end
  def read_only?
    @readwrite == :"read-only"
  end
  def write_only?
    @readwrite == :"write-only"
  end
  def instance?
    @scope == :instance
  end
  def public?
    @access == :public
  end
  def private?
    @access == :private
  end
  def protected?
    @access == :protected
  end
  def inspect
    "<#{self.class} '#{@name}'>"
  end
end

class OMD::Method
  include OMD::Documentable
  include OMD::OwnedByClass
  include OMD::DefinedInSingleFile
  
  attr_reader :name, :parameters, :return_values
  attr_accessor :source_file, :block, :scope, :package

  def initialize( name )
    @name = name
    @parameters = []
    @return_values = []
    @block = :forbidden  # schema default
    @scope = :instance   # schema default
  end
  
  def returns_multiple?
    @return_values.length > 1
  end
  
  def return_value
    @return_values.first
  end
  
  def self.from_omd_xml( element )
    omd_method = new( element.attributes[ 'name' ] )
    omd_method.add_documentation_from_omd_xml( element )
    omd_method.find_source_file_from_omd_xml( element )

    if element.attributes[ 'scope' ]
      omd_method.scope = element.attributes[ 'scope' ].to_sym
    end
    if element.attributes[ 'block' ]
      omd_method.block = element.attributes[ 'block' ].to_sym
    end

    element.each_element( 'Parameter' ){ |param_element|
      omd_attribute = OMD::Method::Parameter.from_omd_xml( param_element )
      omd_method.parameters << omd_attribute
      omd_attribute.owning_method = omd_method
    }
    
    element.each_element( 'ReturnValue' ){ |rv_element|
      omd_rv = OMD::Method::ReturnValue.from_omd_xml( rv_element )
      omd_method.return_values << omd_rv
      omd_rv.owning_method = omd_method
    }
    
    omd_method.package = $current_package
    
    omd_method
  end

  def instance?
    @scope == :instance
  end

  def inspect
    "<#{self.class} '#{@name}'>"
  end
end

class OMD::Method::Parameter
  include OMD::Documentable
  include OMD::OwnedByMethod
  attr_reader :name
  attr_accessor :type, :required, :default
  def initialize( name )
    @name          = name
    @required      = true          # Schema default
  end
  def required?
    @required
  end
  def self.from_omd_xml( element )
    omd_param = new( element.attributes[ 'name' ] )
    omd_param.add_documentation_from_omd_xml( element )
    omd_param.type    = element.attributes[ 'type' ]
    omd_param.default = element.attributes[ 'default' ]
    omd_param.required = !(%w|0 false|.include?( element.attributes[ 'required' ] ))
    
    omd_param
  end
end

class OMD::Method::ReturnValue
  include OMD::Documentable
  include OMD::OwnedByMethod
  attr_accessor :type
  def self.from_omd_xml( element )
    omd_rv = new( )
    omd_rv.add_documentation_from_omd_xml( element )
    omd_rv.type = element.attributes[ 'type' ]
    
    omd_rv
  end
end

if __FILE__==$0
  require 'test/unit'
  class TestOMD < Test::Unit::TestCase
    def setup
      $owl_xml = IO.read( "owlscribble.omd" )
      $dummy_xml = IO.read( "testlang.omd" )
    end
    def test_a_version_parser
      v0_1     = OMD.version_string_value "0.1"
      v1       = OMD.version_string_value "1" 
      v1_0     = OMD.version_string_value "1.0" 
      v1_0_0   = OMD.version_string_value "1.0.0"
      v1_0_0_0 = OMD.version_string_value "1.0.0.0"
      v1_9     = OMD.version_string_value "1.9"
      v1_999   = OMD.version_string_value "1.999"
      v1_999_0 = OMD.version_string_value "1.999.0"
      v2       = OMD.version_string_value "2"
      v2_0     = OMD.version_string_value "2.0"
      v2_0_0   = OMD.version_string_value "2.0.0"
      v2_0_1   = OMD.version_string_value "2.0.1"
      v10      = OMD.version_string_value "10"
      v10_0    = OMD.version_string_value "10.0"

      v10_0_0_0_97823 = OMD.version_string_value "10.0.0.0.97823"
      v10_0_0_1_0     = OMD.version_string_value "10.0.0.1.0"
      v11_0_0         = OMD.version_string_value "11.0.0"
      
      assert_equal( v1, v1_0 )
      assert_equal( v1, v1_0_0 )
      assert_equal( v1, v1_0_0_0 )
      assert( v0_1 < v1, "0.1 should be less than 1" )
      assert( v1 < v1_9, "1 should be less than 1.9" )
      assert( v1_9 < v1_999, "1.9 should be less than 1.999")
      assert_equal( v1_999, v1_999_0 )
      assert( v1_999 < v2, "1.999 should be less than 2" )
      assert_equal( v2, v2_0 )
      assert_equal( v2, v2_0_0 )
      assert( v0_1 < v2, "0.1 should be less than 2" )
      assert( v1 < v2, "1 should be less than 2" )
      assert( v2 < v2_0_1, "2 should be less than 2.0.1" )
      assert( v2_0_1 < v10, "2.0.1 should be less than 10" )
      assert_equal( v10, v10_0 )
      assert( v10 < v10_0_0_0_97823, "10 should be less than 10.0.0.0.97823")
      assert( v10_0_0_0_97823 < v10_0_0_1_0, "10.0.0.0.97823 should be less than 10.0.0.1.0")
      assert( v10_0_0_0_97823 < v11_0_0, "10.0.0.0.97823 should be less than 11.0.0")
    end
    def test_g1_parser
      packages = OMD.load( $owl_xml )
      
      assert_equal( 2, packages.length )
      assert_equal( 6, OMD::CLASSES.length )

      owl_package = OMD::PACKAGES[ 'net.phrogz.owlscribble' ]
      
      assert_equal( 
        packages.find{ |package| package.name=="OWLScribble" },
        owl_package
      )
      assert_equal(
        %w| OWLScribble
            String TagTreeScanner
            TagTreeScanner::Tag
            TagTreeScanner::TagFactory
            TagTreeScanner::TextNode
        |,
        OMD::CLASSES.values.map{ |c| c.name }.sort
      )
      owl_class = OMD::CLASSES[ 'OWLScribble' ]
      assert_equal( :class, owl_class.type )
      assert_equal(
        OMD::CLASSES[ 'TagTreeScanner' ],
        owl_class.parent_classes.first
      )
      assert_equal( 1, owl_class.parent_classes.size )
      
      owl_version = owl_class.constants.select{ |c| c.name == 'VERSION' }.last
      assert_equal( OMD::CLASSES[ 'String' ], owl_version.type )
      assert_equal( "0.9.1", owl_version.value )
      assert_equal( "owlscribble.rb", owl_version.source_file )

      assert(
        owl_package.source_files.include?( "owlscribble.rb" ),
        "OWLScribble package should list 'owlscribble.rb' as a source file."
      )
      assert(
        owl_class.source_files.include?( "owlscribble.rb" ),
        "OWLScribble class should list 'owlscribble.rb' as a source file."
      )
      assert_equal( 1, owl_class.own_constants.length, "OWLScribble has 1 constant." )
      assert_equal( 5, owl_class.own_attributes.length, "OWLScribble has 5 attributes." )
      assert_equal( 5, owl_class.own_methods.length, "OWLScribble has 5 methods." )
      
      toc = owl_class.attributes.find{ |a| a.name == 'table_of_contents' }
      assert_equal(
        OMD::CLASSES[ 'TagTreeScanner::Tag' ],
        toc.type
      )
    end
    
    def test_g2_parser
      packages = OMD.load( $dummy_xml )
      assert_equal( 3, packages.length )
      trucks = OMD::PACKAGES[ 'net.phrogz.trucks' ]
      
      assert_equal( 11, OMD::CLASSES.length )
      truck        = OMD::CLASSES[ 'Truck' ]
      dumper       = OMD::CLASSES[ 'DumpTruck' ]
      osha         = OMD::CLASSES[ 'OSHAEquipment' ]
      vehicle      = OMD::CLASSES[ 'Vehicle' ]
      construction = OMD::CLASSES[ 'ConstructionEquipment' ]
      [ truck, dumper ].each{ |klass| assert_equal( :class, klass.type ) }
      [ osha, vehicle, construction ].each{ |mod| assert_equal( :module, mod.type ) }
      assert_equal( 3, dumper.parent_classes.length )
      assert_equal( truck, dumper.parent_classes[0] )
      assert_equal( vehicle, dumper.parent_classes[1] )
      assert_equal( construction, dumper.parent_classes[2] )
      
      assert_equal( 1, dumper.own_methods.length )
      assert_equal( 4, dumper.inherited_methods.length )
      assert_equal( 5, dumper.methods.length )
      assert_equal( 0, dumper.own_constants.length )
      assert_equal( 1, dumper.inherited_constants.length )
      assert_equal( 1, dumper.constants.length )
    end

  end
end