#!/usr/bin/env ruby

require 'omd'
require 'fileutils'
require 'erb'
module OMD::HTML
  INCLUDES  = "_includes"
  TEMPLATES = "_templates"
  INDEX_TEMPLATE     = "index.rhtml"
  PACKAGE_TEMPLATE   = "package.rhtml"
  CLASS_TEMPLATE     = "class.rhtml"
  CONSTANT_TEMPLATE  = "constant.rhtml"
  ATTRIBUTE_TEMPLATE = "attribute.rhtml"
  METHOD_TEMPLATE    = "method.rhtml"
  PARAMETER_TEMPLATE = "parameter.rhtml"

  def self.write_omd( output_directory )
    dirs_created = 0
    files_created = 0

    Dir.mkdir( output_directory )
    dirs_created += 1
    
    # (Re)copy include files, in case they changed
    includes_dest = File.join( output_directory, INCLUDES )
    Dir.mkdir( includes_dest )
    dirs_created += 1
    Dir.entries( INCLUDES ).each{ |file|
    	next if file =~ /^\./
    	FileUtils.copy( File.join( INCLUDES, file) , includes_dest )
      files_created += 1
    	# Ensure that the newly-copied file is read-write
    	File.chmod( 0644, File.join( includes_dest, file ) )
    }

    require File.join( TEMPLATES, 'template_helpers.rb' )
    index_template     = IO.read( File.join( TEMPLATES, INDEX_TEMPLATE ) )
    package_template   = IO.read( File.join( TEMPLATES, PACKAGE_TEMPLATE ) )
    class_template     = IO.read( File.join( TEMPLATES, CLASS_TEMPLATE ) )
    constant_template  = IO.read( File.join( TEMPLATES, CONSTANT_TEMPLATE ) )
    attribute_template = IO.read( File.join( TEMPLATES, ATTRIBUTE_TEMPLATE ) )
    method_template    = IO.read( File.join( TEMPLATES, METHOD_TEMPLATE ) )
    parameter_template = IO.read( File.join( TEMPLATES, PARAMETER_TEMPLATE ) )
    
    # Create the index page
		file_name = File.join( output_directory, "index.html" )
		html = ERB.new( index_template ).result( binding )
		File.open( file_name, "w" ){ |out_file| out_file << html }
		files_created += 1
    
    OMD::PACKAGES.each{ |package_id, omd_asset|
  		file_name = File.join( output_directory, omd_asset.file_path )
  		html = ERB.new( package_template ).result( binding )
  		File.open( file_name, "w" ){ |out_file| out_file << html }
  		files_created += 1
    }
    
    OMD::CLASSES.each{ |name, omd_class|
      omd_asset = omd_class
      Dir.mkdir( File.join( output_directory, omd_asset.directory ) )

  		file_name = File.join( output_directory, omd_asset.file_path )
  		html = ERB.new( class_template ).result( binding )
  		File.open( file_name, "w" ){ |out_file| out_file << html }
  		files_created += 1

  		omd_class.own_constants.each{ |omd_asset|
    		file_name = File.join( output_directory, omd_asset.file_path )
    		html = ERB.new( constant_template ).result( binding )
    		File.open( file_name, "w" ){ |out_file| out_file << html }
    		files_created += 1
  		}

  		omd_class.own_attributes.each{ |omd_asset|
    		file_name = File.join( output_directory, omd_asset.file_path )
    		html = ERB.new( attribute_template ).result( binding )
    		File.open( file_name, "w" ){ |out_file| out_file << html }
    		files_created += 1
  		}
  		
  		omd_class.own_methods.each{ |omd_asset|
    		file_name = File.join( output_directory, omd_asset.file_path )
    		html = ERB.new( method_template ).result( binding )
    		File.open( file_name, "w" ){ |out_file| out_file << html }
    		files_created += 1
  		}
    }
    
    [ dirs_created, files_created ]
  end
end

if __FILE__==$0
  USAGE = <<ENDUSAGE
Usage:
   omd [-f] source.omd [-d | dest_directory]
   -f,--force   force overwriting of an existing directory
   -d,--derive  derive a non-unique directory from the source file name
                (ignored if a dest_directory is supplied)
   -h,--help    show some help examples
ENDUSAGE
  
EXAMPLES = <<ENDEXAMPLES

Examples:
   omd bindings.omd
   # Will create a directory named "bindings_20080115a",
   # using the current date and a unique trailing letter.

   omd bindings.omd -d
   # Will create a directory named "bindings";
   # will not overwrite such a directory if it exists.

   omd bindings.omd -d -f
   # Will create a directory named "bindings",
   # overwriting an existing one if it exists.

   omd player_runtime.omd PlayerRuntime
   # Will create a directory named "PlayerRuntime";
   # will not overwrite if that directory already exists.

   omd -f player_runtime.omd PlayerRuntime
   # Will create a directory named "PlayerRuntime",
   # first clearing it erasing if it already exists.
ENDEXAMPLES
  
  ARGS = {}
  UNFLAGGED_ARGS = [ :source_file, :output_directory ]
  next_arg = UNFLAGGED_ARGS.first
  ARGV.each{ |arg|
     case arg
       when '-f','--force'
         ARGS[:force] = true
       when '-d','--derive'
         ARGS[:derive] = true
       when '-h','--help'
         ARGS[:help] = true
       else
         if next_arg
           ARGS[next_arg] = arg
           UNFLAGGED_ARGS.delete( next_arg )
         end
         next_arg = UNFLAGGED_ARGS.first
     end
  }
  
  if ARGS[:help] or !ARGS[:source_file]
     puts USAGE
     puts EXAMPLES if ARGS[:help]
     exit
  end
  
  OMD.load( IO.read( ARGS[:source_file] ) )
  unless out_dir = ARGS[ :output_directory ]
    filebase = File.basename( ARGS[:source_file], File.extname( ARGS[:source_file] ) )
    if ARGS[:derive]
      out_dir = filebase
    else
      out_dir = "#{filebase}_#{Time.now.strftime "%Y%m%d"}a"
      out_dir.succ! while File.exist?( out_dir )
    end
  end
  
  if File.exist?( out_dir )
    raise "Output directory #{out_dir} exists, and --force option not specified." unless ARGS[:force]
    FileUtils.rm_rf( out_dir )
  end

  start_time = Time.now
  directories, files = OMD::HTML.write_omd( out_dir )
  end_time = Time.now
  
  puts "#{directories} directories and #{files} files created in %.3f seconds." % (end_time - start_time)
end