module OMD::HTML::Linkable
  def as_link( name_prefix="", link_text=nil )
    if link_text
      "<a href='#{File.join( $relative_path, file_path )}' title='#{short_attr overview}'>#{link_text}</a>"
    else
      "<a href='#{File.join( $relative_path, file_path )}' title='#{short_attr overview}'>#{name_prefix}#{name}</a>"
    end
  end
end

class OMD::Package
  include OMD::HTML::Linkable
  def file_path
    "#{id.windows_safe}.html"
  end
end

class OMD::Class
  include OMD::HTML::Linkable
  def directory
    name.windows_safe
  end
  def file_path
    File.join( directory, 'index.html' )
  end
end

class OMD::Constant
  include OMD::HTML::Linkable
  def file_path
    File.join( owning_class.directory, "c_#{name.windows_safe}.html" )
  end
end

class OMD::Attribute
  include OMD::HTML::Linkable
  def file_path
    File.join( owning_class.directory, "a#{:c unless instance?}_#{name.windows_safe}.html" )
  end
end

class OMD::Method
  include OMD::HTML::Linkable
  def file_path
    #TODO: What if multiple methods have the same name? Disambiguate with some index...
    File.join( owning_class.directory, "m#{:c unless instance?}_#{name.windows_safe}.html" )
  end
end

class OMD::Method::Parameter
  include OMD::HTML::Linkable
  def file_path
    File.join( owning_method.owning_class.directory, "m#{:c unless owning_method.instance?}_#{owning_method.name.windows_safe}_param#{owning_method.parameters.index(self)}.html" )
  end
  def to_html
    pname = docs.blank? ? name : as_link
    "<span class='param#{' optional' unless required?}'>#{pname}#{"=#{default}" if default}</span>"
  end
end

class String
  def windows_safe
    self.tr( '/\\:"*?<>|', '' ).tr( '.', '_' )
  end
  def as_link( *a )
    self
  end
  def blank?
    empty? || ( /\A\s+\Z/ =~ self )
  end
end

module OMD::Documentable
  def pretty_docs
    @pretty_docs ||= make_pretty( docs )
  end
  def pretty_overview
    @pretty_overview ||= make_pretty( overview )
  end
  def make_pretty( str )
    if str
      str = CGI.escapeHTML( str )
      str.gsub!( /%(\S+)%/ ){
        code_ref = $1
        omd_asset = html_resolve_reference( code_ref )
        omd_asset.as_link( nil, code_ref )
      }
      str.gsub!( /\{\{\{(.+?)\}\}\}/, '<code>\\1</code>' )
      str.gsub!( %r{((?:ftp|http)://\S+)}, '<a href="\\1">\\1</a>' )
      str
    end
  end
  
  def create_reference_lookup( root_is_more_important_than_self=false )
    lookup = OMD::CLASSES.dup
    klass_scope = case self
      when OMD::Class
        self
      when OMD::OwnedByClass, OMD::OwnedByMethod
        owning_class
    end
    
    # Not true for Packages, or Parameters
    if klass_scope
      const_lookup = Hash[ *klass_scope.own_constants.map{  |o| [o.name,o] }.flatten ]
      attrs_lookup = Hash[ *klass_scope.own_attributes.map{ |o| [o.name,o] }.flatten ]
      meths_lookup = Hash[ *klass_scope.own_methods.map{    |o| [o.name,o] }.flatten ]
      lookup.update( const_lookup ).update( attrs_lookup ).update( meths_lookup )
    end
    
    if self.is_a?( OMD::OwnedByMethod )
      params_lookup = Hash[ *self.owning_method.parameters.map{ |o| [o.name,o] }.flatten ]
      lookup.update( params_lookup )
    end
    
    lookup.update( OMD::CLASSES ) if root_is_more_important_than_self
    
    lookup
  end
  
  def html_resolve_reference( code_ref )
    pieces = code_ref.scan( /\w+/ )
    
    omd_asset = nil
    lookup = create_reference_lookup( true )
    pieces.each{ |piece|
      # TODO: try to singularize plural references as a fallback
      return code_ref unless omd_asset = lookup[ piece ]
      lookup = omd_asset.create_reference_lookup
    }
    omd_asset
  end
end

=begin
require 'rubygems'
require 'RedCloth', '= 3.0.4'

module OMD::Documentable
		def overview_pretty
			make_pretty( @overview ).gsub( %r{</?p>}, '' )
		end

		def description_pretty
			make_pretty( @description )
		end
end

class String
	def strip_indent
		if indent = /^(?:\s*[\n\r])?([ \t]*)(?=\S)/.match( self )
			self.strip.gsub( /^#{indent[1]}/, '' )
		else
			self.strip
		end
	end
end

# Use RedCloth to convert markup to HTML, and add html links within the framework
def make_pretty( str )
	return '' unless str
	str = RedCloth.new( str.strip_indent.gsub( /\t/, '  ' ) ).to_html
	info = {}
	{
		:classes => $framework.known_classes,
		:methods => $framework.methods,
		:properties => $framework.properties
	}.each{ |k,a|
		info[ k ] = {
			:names => a.collect{ |o| o.name unless o.name == 'a' }.compact.join( '|' ),
			:by_name => a.to_hash{ |o| [o.name,o] }
		}
	}

	str.gsub!( /((#{info[:classes][:names]})[.:])?\b(#{info[:methods][:names]})\(\)/ ){ |orig|
		matching_methods=info[:methods][:by_name][$3]
		raise "\n\n$1:#{$1}, $2:#{$2}, $3:#{$3}\n\n" unless matching_methods
		if matching_methods.length == 1
			meth = matching_methods.first
		else
			meth = matching_methods.find{ |m| m.script_class.name == $2 }
		end
		if meth
			"#{$1}<code><a href='#{meth.file_path}'>#{meth.name}()</a></code>"
		else
			orig #Ambiguous method name - no link
		end
	}

# Properties were matching too many common words like 'type'
	# str.gsub!( /((#{info[:classes][:names]})\.)?\b(#{info[:properties][:names]})\b/ ){ |orig|
	# 	matching_props=info[:properties][:by_name][$3]
	# 	if matching_props.length == 1
	# 		prop = matching_props.first
	# 	else
	# 		prop = matching_props.find{ |prop| prop.script_class.name == $2 }
	# 	end
	# 	if prop
	# 		"#{$1}<code><a href='#{prop.file_path}'>#{prop.name}</a></code>"
	# 	else
	# 		orig
	# 	end
	# }

	str.gsub!( /\b(#{info[:classes][:names]})\b(?!\.html|_[mp]|>)/ ){ |orig|
		sclass=info[:classes][:by_name][$1].first
		"<code><a href='#{sclass.file_path}'>#{orig}</a></code>"
	}

	str
end
=end

require 'cgi'
def short_attr( str )
  if str
    str = CGI.escapeHTML( str ).gsub( "'", "&apos;" )
    str.gsub!( /[{}]{3}/, '' )
    str.gsub!( /%(\S+)%/, '\\1' )
    str
  end
  
  # TODO: shorten
end

def a( str )
  if str =~ /^[aeiouy]/i
    "an #{str}"
  else
    "a #{str}"
  end
end

class NilClass
  def blank?
    true
  end
  def as_link(*a)
    self
  end
end
